/*
 *	epos/src/options.cc
 *	(c) 1996-01 geo@cuni.cz
 *
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License in doc/COPYING for more details.
 *
 */

#include "common.h"

#include "slab.h"
SLABIFY(configuration, cfg_slab, 2, shutdown_cfgs)

#define CMD_LINE_OPT	"-"
#define CMD_LINE_VAL	'='

#define INDEXER		"@"

#ifdef HAVE_UNISTD_H
	#include <unistd.h>
#endif

inline void *memdup(void *p, int size)
{
	if (!p) {
		return NULL;
	}
	void *r = xmalloc(size);
	memcpy(r,p,size);
	return r;
}

configuration proto_cfg;

configuration *cfg = &proto_cfg;

static_configuration proto_scfg;
static_configuration *scfg = &proto_scfg;


/*
 *	cow - this routine should detect whether a piece of memory used
 *		for global/language/voice configuration is shared with
 *		supposedly logical copies, and, if so, copy it physically
 *		before changing it. A nice space-saving technique.
 *
 *	cow_claim - claims all current configuration
 *	cow_unclaim - releases all current configuration; cowabilia without other
 *		previous claimers are freed
 *
 *	??Semantics: cow == number of claimers minus one
 *
 *	The implementation is terrible (though fast). If I knew how to design this
 *	correctly, I would do that. I have done it twice wrong and it grew moderately
 *	complex during the process of fixing.
 *
 */

void cow(cowabilium **p, int size, int extraoffset, int extrasize)
{
	cowabilium *src = *p;
	cowabilium **ptr = p;

	if (!*p) {
//		if (p == (cowabilium **)&this_lang) shriek(814, "Language-specific option not supported on the cmd line");
//		if (p == (cowabilium **)&this_voice) shriek(814, "Voice-specific option not supported on the cmd line");
		shriek(861, "cow()ing a NULL");
	}

	if (src->cow) {
//		*ptr = (cowabilium *)xmalloc(size);
//		printf("Copying %p(%d) to %p\n", src, src->cow, *ptr);
//		memcpy(*ptr, src, size);

		*ptr = (cowabilium *)memdup(src, size);
		(*ptr)->cow = 0;
		(*ptr)->parent = src;
		src->cow--;

		if (extraoffset) {
			*(void **)((char *)*ptr + extraoffset)
				= memdup(*(void **)((char *)src + extraoffset), extrasize);
		}
	}
}

#define  LANGS_OFFSET   ((int)&((configuration *)NULL)->langs)
#define  LANGS_LENGTH   ((*cfg)->n_langs * sizeof(void *))

void cow_configuration(configuration **cfg)
{
	cow((cowabilium **)cfg, sizeof(configuration), LANGS_OFFSET, LANGS_LENGTH);
}

void cow_claim()
{
//	printf("Claiming %p, was %d\n", cfg, cfg->cow);
	for (int i=0; i < cfg->n_langs; i++) {
		lang *l = cfg->langs[i];
//		printf("   lang %s(%p) was %d\n", l->name, l, l->cow);
		l->cow++;
		for (int j=0; j < l->n_voices; j++) {
//			printf("      voice %s(%p) was %d\n", l->voices[j]->name, l->voices[j], l->voices[j]->cow);
			l->voices[j]->cow++;
		}
	}
	cfg->cow++;
}

void cow_unstring(cowabilium *p, epos_option *opts)
{
	for (epos_option *o = opts; o->optname; o++) {
		if (o->opttype != O_STRING || o->per_level) continue;
		char **to_free = (char **)((char *)p + o->offset);
		if (*to_free && (*(char **)((char *)p->parent + o->offset) != *to_free
				|| p->parent == p)) {
//			printf("Freeing %s\n", o->optname);
			free(*to_free);
			*to_free = NULL;
		}
	}
}

static void cow_unsynth(voice *v)
{
	voice *par = (voice *)v->parent;
	if (v->syn != par->syn) delete v->syn;
	if (v->segment_names != par->segment_names) unclaim(v->segment_names);
	if (v->sl != par->sl) free(v->sl);
}

static inline void cow_free(cowabilium *p, epos_option *opts, void *extra)
{
	cow_unstring(p, opts);
	if (opts == voiceoptlist) cow_unsynth((voice *)p);
	free(p);
	if (extra) free(extra);
}

extern epos_option optlist[];

void cow_unclaim(configuration *that_cfg)
{
//	printf("Unclaiming %p, was %d\n", that_cfg, that_cfg->cow);
	for (int i=0; i < that_cfg->n_langs; i++) {
		lang *l = that_cfg->langs[i];
//		printf("   lang %s(%p) was %d\n", l->name, l, l->cow);
		for (int j=0; j < l->n_voices; j++) {
//			printf("      voice %s(%p) was %d\n", l->voices[j]->name, l->voices[j],l->voices[j]->cow);
			if (!l->voices[j]->cow--) cow_free(l->voices[j], voiceoptlist, NULL);
			/* * * fortunately, soft options are never strings * * */
		}
		if (!l->cow--) cow_free(l, langoptlist, l->voices);
	}
	if (!that_cfg->cow--) cow_free(that_cfg, optlist, that_cfg->langs), that_cfg = NULL;
}

void cow_catharsis()
{
	cow_unclaim(cfg);
//	cfg = &proto_cfg;
//	cow_claim();
}

#define CONFIG_INITIALIZE
inline configuration::configuration() : cowabilium()
{
	#include "options.lst"

	n_langs = 0;
	langs = NULL;

	stdshriek = stderr;
	stddbg = stdout;

	current_stream = NULL;
}

configuration::~configuration()
{
	cow_unstring(this, optlist);
}

#define EO {NULL, O_INT, OS_CFG, A_PUBLIC, A_PUBLIC, false, false, -1},
#define TWENTY_EXTRA_OPTIONS	EO EO EO EO EO EO EO EO EO EO EO EO EO EO EO EO EO EO EO EO 

// configuration master_cfg;

#define CONFIG_DESCRIBE
epos_option optlist[]={
        #include "options.lst"

	{"C:language" + 2, O_LANG, OS_CFG, A_PUBLIC, A_PUBLIC, false, false, 0},
	{"", O_INT, OS_CFG, A_PUBLIC, A_PUBLIC, false, false, -3},
	TWENTY_EXTRA_OPTIONS
	TWENTY_EXTRA_OPTIONS
	TWENTY_EXTRA_OPTIONS
	TWENTY_EXTRA_OPTIONS
	TWENTY_EXTRA_OPTIONS
	TWENTY_EXTRA_OPTIONS
	{NULL, O_INT, OS_CFG, A_PUBLIC, A_PUBLIC, false, false, -2}
};

#undef EO
#undef TWENTY_EXTRA_OPTIONS


#define CONFIG_STATIC_INITIALIZE
inline static_configuration::static_configuration()
{
	#include "options.lst"
}

#define EO {NULL, O_INT, OS_STATIC, A_PUBLIC, A_PUBLIC, false, false, -1},
#define TWENTY_EXTRA_OPTIONS	EO EO EO EO EO EO EO EO EO EO EO EO EO EO EO EO EO EO EO EO 

#define CONFIG_STATIC_DESCRIBE
epos_option staticoptlist[]={
	#include "options.lst"
	{"", O_INT, OS_STATIC, A_PUBLIC, A_PUBLIC, false, false, -3},
	TWENTY_EXTRA_OPTIONS
	TWENTY_EXTRA_OPTIONS
	TWENTY_EXTRA_OPTIONS
	{NULL, O_INT, OS_STATIC, A_PUBLIC, A_PUBLIC, false, false, -2}
};

#undef EO
#undef TWENTY_EXTRA_OPTIONS


hash_table<char, epos_option> *option_set = NULL;

void configuration::shutdown()
{
	int i;
	for (i=0; i < n_langs; i++) delete langs[i];
	free(langs);
	langs = NULL;
//	this_lang = NULL; this_voice = NULL;
	n_langs = 0;
}

/*
 * Some string trickery below. option->name is initialised (in options.lst)
 *	to   "X:" "string"+2. In C, it means the same as "X:string"+2,
 *	which means "string" with "X:" immediately preceding it.
 * Now we add to our option_set both normal option names ("string")
 *	and their prefixed spellings ("X:string", for X being either
 *	S, C, L or V for static/global/language/voice configuration items,
 *	respectively), eating up minimum memory space only.
 */

inline void put_into_option_set(epos_option *o)
{
	if (*o->optname) {
		option_set->add(o->optname, o);
		option_set->add(o->optname - 2, o);
	}
}

void make_option_set()
{
	epos_option *o;

	if (option_set) return;

	option_set = new hash_table<char, epos_option>(300);
	option_set->dupkey = option_set->dupdata = 0;

	for (o = staticoptlist; o->optname; o++)put_into_option_set(o);
	for (o = optlist; o->optname; o++) 	put_into_option_set(o);
	for (o = langoptlist; o->optname; o++)	put_into_option_set(o);
	for (o = voiceoptlist; o->optname; o++)	put_into_option_set(o);

	option_set->remove("languages");	// FIXME
	option_set->remove("voices");
}

void restrict_options()
{
	epos_option *o;

	text *t = new text(scfg->restr_file, scfg->ini_dir, "", NULL, true);
	if (!t->exists()) {
		delete t;
		return;		/* if no restrictions file, ignore it */
	}
	char *line = (char *)xmalloc(scfg->max_line);
	while (t->getline(line)) {
		char *status = split_string(line);
		o = option_set->translate(line);
		if (!o) {
			if (cfg->paranoid)
				shriek(812, "Typo in %s:%d\n", t->current_file, t->current_line);
			continue;
		}
		o->readable = o->writable = A_NOACCESS;
		ACCESS level = A_PUBLIC;
		for (char *p = status; *p; p++) {
			switch(*p) {
				case 'r': o->readable = level; break;
				case 'w': o->writable = level; break;
				case '#': level = A_ROOT; break;
				case '$': level = A_AUTH; break;
				default : shriek(812, "Invalid access rights in %s line %d",
						t->current_file, t->current_line);
			}
		}
	}
	free(line);
	delete t;
}

static void free_all_options(epos_option *optlist, cowabilium *whence)
{
	epos_option *o;
	for (o = optlist; o->offset != -2; o++) ;
	for ( ; o->offset != -3; o--) ;
	for ( o++; o->offset != -2; o++) {
		if (o->offset == -1) continue;
		if (o->opttype == O_STRING && whence) {
			if (*(char **)((char *)whence + o->offset)) {
				free(*(char **)((char *)whence + o->offset));
				*(char **)((char *)whence + o->offset) = NULL;
			}
		}
		if (o->optname) {
			D_PRINT(1, "Freeing an extra option %s\n", o->optname - 2);
			option_set->remove(o->optname);
			option_set->remove(o->optname - 2);
			free((char *)o->optname - 2);
			o->optname = NULL;
		}
		o->offset = -1;
	}
}

void free_all_options()
{
	free_all_options(voiceoptlist, NULL);
	free_all_options(langoptlist, NULL);
	free_all_options(optlist, cfg);
	free_all_options(staticoptlist, (cowabilium *)scfg);
}

epos_option *alloc_option(epos_option *optlist, OPT_STRUCT os)
{
	DBG(1, int j = 0; epos_option *p; for (p = optlist; p->offset != -2; p++) if (p->offset == -1) j++; fprintf(STDDBG, "Allocating an extra option, level %d out of %d free options\n", os, j);)

	epos_option *o;
	for (o = optlist; ; o++) {
		if (o->offset == -1) return o;
		if (o->offset == -2) shriek(864, "No more options to allocate, add extra options");
	}
	return NULL;	// to keep the compiler happy
}

#define MAX_TOTAL_OPTION_NAME 64

void alloc_level_options(epos_option *optlist, OPT_STRUCT os, void *base, int levnum, const char *levname)
{
	char b[MAX_TOTAL_OPTION_NAME];

	epos_option *o;
	for (o=optlist; o->optname; o++) {
		if (o->per_level) {
			epos_option *p = alloc_option(optlist, os);
			b[0] = OPT_STRUCT_PREFIX[os];
			b[1] = ':';
			strcpy(b+2, o->optname);
			strcat(b+2, INDEXER);
			strcat(b+2, levname);
			p->optname = strdup(b) + 2;
			p->opttype = o->opttype;
			p->readable = o->readable;
			p->writable = o->writable;
			option_set->add(p->optname - 2, p);
			option_set->add(p->optname, p);

			int size;
			switch (o->opttype) {
				case O_STRING:	size = sizeof(char *); break;
				case O_INT:	size = sizeof(int); break;
				case O_BOOL:	size = sizeof(bool); break;
				default: shriek(861, "strange type size_of-ed"); break;
			}
			p->offset = o->offset + levnum * size;
			if (base)
				memcpy((char *)base + p->offset, (char *)base + o->offset, size);
					// CHECKME cow
		}
	}
}


#define LEVNAME_SEGMENT	"segment"
#define LEVNAME_PHONE	"phone"
#define LEVNAME_TEXT	"text"

static inline const char *pre_set_unit_levels(const char *value)
{
	if ((str2enum(LEVNAME_SEGMENT, value, U_DEFAULT) == U_ILL) ||
		(str2enum(LEVNAME_PHONE, value, U_DEFAULT) == U_ILL) ||
		(str2enum(LEVNAME_TEXT, value, U_DEFAULT) == U_ILL))
			return NULL;		// required level missing

	char b[MAX_TOTAL_OPTION_NAME];
	int n;

	for (n=0; enum2str(n, value); n++) ;
	if (n >= UNIT_MAX) return NULL;		// too many levels
	for (int i=0; i<n; i++) {
		strcpy(b, enum2str(i, value));
		for (int j=0; j<n; j++) {
			if (i!=j && !strcmp(b, enum2str(j, value))) 
				return NULL;	// two level names collide
		}
	}
	return value;
}

static inline void post_set_unit_levels(const char *)
{
	scfg->segm_level = str2enum(LEVNAME_SEGMENT, scfg->unit_levels, U_DEFAULT);
	scfg->phone_level = str2enum(LEVNAME_PHONE, scfg->unit_levels, U_DEFAULT);
	scfg->text_level = str2enum(LEVNAME_TEXT, scfg->unit_levels, U_DEFAULT);
	scfg->default_scope = str2enum("word", scfg->unit_levels, U_DEFAULT);
	if (scfg->default_scope == U_ILL) scfg->default_scope = scfg->text_level;
	scfg->default_target = scfg->phone_level;

	int n;
	for (n=0; enum2str(n, scfg->unit_levels); n++) ;

	for (int i=0; i<n; i++) {
		const char *tmp = enum2str(i, scfg->unit_levels);
		alloc_level_options(staticoptlist, OS_STATIC, scfg, i, tmp);
		alloc_level_options(optlist, OS_CFG, cfg, i, tmp);
		alloc_level_options(langoptlist, OS_LANG, NULL, i, tmp);
	}
}

static inline const char *pre_base_dir(const char *value)
{
	if (value && value[0] != SLASH) {
		size_t size = scfg->scratch ? scfg->scratch : 64;
		char *abs = (char *)xmalloc(size);
		while (!getcwd(abs, size - 1))
			size *= 2, abs = (char *)xrealloc(abs, size);
		while (strlen(abs) + strlen(value) + 1 >= size)
			size *= 2, abs = (char *)xrealloc(abs, size);
		char *tmp = abs + strlen(abs);
		*tmp++ = SLASH;
		strcpy(tmp, value);
		char *abs2 = FOREVER(strdup(abs));
		free(abs);
		return abs2;
	} else return value;
}

static inline const char *pre_init_f(const char *value)
{
	int v;
	if (!sscanf(value,"%d",&v)) return value;
	if (v < 1) return NULL;
	else return value;
}

/*
 *	invoke_set_action gets invoked twice for options whose action bit
 *		is set.  The pre-set call receives and returns a value,
 *		the post-set call receives a NULL and returns NULL.
 *
 *		If a pre-set call returns a NULL, the option will not
 *		be changed and no post-set call occurs.  Otherwise the
 *		value returned is used instead of the suggested one.
 *
 *		The value returned by invoke_set_action (if different
 *		from the argument) can and should be allocated using
 *		scratch.  invoke_set_action must not be called recursively.
 *
 *		All options have initially the action bit set, and
 *		invoke_set_action shall disable the action bit for
 *		options it doesn't handle specially.
 */

const char *invoke_set_action(epos_option *o, const char *value)
{
	const char *on = strchr(o->optname, ':') ? strchr(o->optname, ':') + 1 : o->optname;
	if (!strcmp(on, "base_dir")) {
		if (value) return pre_base_dir(value);
		else return NULL;
	}
	if (!strcmp(on, "init_f")) {
		if (value) return pre_init_f(value);
		else return NULL;
	}
	if (!strcmp(on, "unit_levels")) {
		if (value) {
			return pre_set_unit_levels(value);
		} else {
			post_set_unit_levels(NULL);
			return NULL;
		}
	}
#ifdef HAVE_WINSOCK
	if (!strcmp(on, "readfs")) {
		if (!value) return NULL;
		if (str2enum(value, BOOLstr, 0) & 1) return NULL;
		else return value;
	}
#endif
	o->action = false;
	return value;
}

epos_option *option_struct(const char *name, hash_table<char, epos_option> *softopts)
{
	epos_option *o;

	if (!option_set) shriek(862, "no config_init()");
	if (!name || !*name) return NULL;
	o = option_set->translate(name);
	if (!o && softopts) o = softopts->translate(name);
	return o;
}

void unknown_option(char *name, char *)
{
	shriek(442, "Never heard about \"%s\", option ignored", name);
}

void parse_cfg_str(char *val)
{
	char *brutto, *netto;
	bool tmp;
	
	tmp=0;
	if (*val == DQUOT && *(brutto = val + strlen(val)-1) == DQUOT && brutto[-1] != ESCAPE) 
		tmp = 1, *brutto = 0;       //If enclosed in double quotes, kill'em.
	for (netto=val, brutto = val + tmp; *brutto; brutto++, netto++) {
		*netto = *brutto;
		if (*brutto == ESCAPE) *netto = esctab->xlat(*++brutto);
	}				//resolve escape sequences
	*netto = 0;
}

template<class T> inline void set_enum_option(epos_option *o, const char *val, const char *list, char *locus)
{
	parse_cfg_str(const_cast<char *>(val));
	T tmp = (T)str2enum(val, list, U_ILL);
	if (tmp == (T)U_ILL)
		shriek(447, "Can't set %s to %s", o->optname, val);
	*(T *)locus = tmp;
}

bool set_option(epos_option *o, const char *val, void *base)
{
	int tmp;
	if (!o) return false;
	if (o->action && !(val = invoke_set_action(o, val))) return false;
	char *locus = (char *)base + o->offset;
	switch(o->opttype) {
		case O_BOOL:
			tmp = str2enum(val, BOOLstr, false);
			if (cfg->paranoid && (!val || tmp == U_ILL)) 
				shriek(447, "%s is used as a boolean value for %s", val, o->optname);
			*(bool *)locus = tmp & 1;
			D_PRINT(1, "Configuration option \"%s\" set to %s\n",
				o->optname,enum2str(*(bool*)locus, BOOLstr));
			break;
		case O_MARKUP:
			set_enum_option<OUT_ML>(o, val, OUT_MLstr, locus);
			D_PRINT(1, "Markup language option set to %i\n",*(int*)locus);
			break;
		case O_SYNTH:
			set_enum_option<SYNTH_TYPE>(o, val, STstr, locus);
			D_PRINT(1, "Synthesis type option set to %i\n",*(int*)locus);
			break;
		case O_CHANNEL:
			set_enum_option<CHANNEL_TYPE>(o, val, CHANNEL_TYPEstr, locus);
			D_PRINT(1, "Channel type option set to %i\n",*(int*)locus);
			break;
		case O_UNIT:
//			if((*(UNIT *)locus=str2enum(val, scfg->unit_levels, U_ILL))==U_ILL) 
//				shriek(447, "Can't set %s to %s", o->optname, val);
			set_enum_option<UNIT>(o, val, scfg->unit_levels, locus);
			D_PRINT(1, "Configuration option \"%s\" set to %d\n",o->optname,*(int *)locus);
			break;
		case O_INT:
//			*(int *)locus=0;
			int v;
			if (!sscanf(val,"%d",&v)) shriek(447, "Unrecognized numeric parameter");
			*(int *)locus = v;
			D_PRINT(1, "Configuration option \"%s\" set to %d\n",o->optname,*(int *)locus);
			break;
		case O_STRING: 
			if (val[0]) parse_cfg_str(const_cast<char *>(val));
			D_PRINT(1, "Configuration option \"%s\" set to \"%s\"\n",o->optname,val);
			if (*(char **)locus) free(*(char**)locus);
			*(char**)locus = strdup(val);	// FIXME: should be forever if monolith etc. (maybe)
			break;
		case O_CHAR:
			if (!val[0]) shriek(447, "Empty string given for a CHAR option %s", o->optname);
			parse_cfg_str(const_cast<char *>(val));
			if (val[1]) shriek(447, "Multiple chars given for a CHAR option %s", o->optname);
			else *(char *)locus=*val;
//			D_PRINT(1, "Configuration option \"%s\" set to \"%s\"\n",o->optname,val);
			break;
		case O_LANG:
			if (!lang_switch(val)) shriek(443, "unknown language");
			break;
		case O_VOICE:
			if (!voice_switch(val)) shriek(443, "unknown voice");
			break;
		case O_CHARSET:
			parse_cfg_str(const_cast<char *>(val));
			if ((tmp = load_charset(val)) == CHARSET_NOT_AVAILABLE)
				shriek(447, "Unknown charset %s", val);
			*(int *)locus = tmp;
			D_PRINT(1, "Charset set to %i\n",*(int*)locus);
			break;
		default: shriek(462, "Bad option type %d", (int)o->opttype);
	}
	if (o->action) invoke_set_action(o, NULL);

	return true;
}

/*
 *	C++ is evil with respect to double indirection.
 *	The casts to cowabilium below are a braindamage,
 *	unless the idea of a typed language is a braindamage.
 *	C++ tries to force us to pass it by reference,
 *	but struct cowabilium should not have formed a basis for
 *	inheritance in the first place.
 *	
 *	The order of cowing cfg, lang and voice, is important.
 */

#define  VOICES_OFFSET  ((int)&((lang *)NULL)->voices)
#define  VOICES_LENGTH  (this_lang->n_voices * sizeof(void *))

bool set_option(epos_option *o, const char *value)
{
	if (!o) return false;
	switch(o->structype) {
		case OS_STATIC: return set_option(o,value, scfg);
		case OS_CFG:	cow_configuration(&cfg);
				return set_option(o, value, cfg);
		case OS_LANG:	cow_configuration(&cfg);
				cow((cowabilium **)&this_lang, sizeof(lang),   VOICES_OFFSET, VOICES_LENGTH);
				return set_option(o, value, this_lang);
		case OS_VOICE:	cow_configuration(&cfg);
				cow((cowabilium **)&this_lang, sizeof(lang),   VOICES_OFFSET, VOICES_LENGTH);
				cow((cowabilium **)&this_voice, sizeof(voice) + (this_lang->soft_options->items * sizeof(void *) >> 1), 0, 0);
				return set_option(o, value, this_voice);
	}
	return false;
}

static inline bool set_option(char *name, const char *value)
{
//	option *o = option_struct(name, NULL);
//	... tvorba indexu ...
//	char *tmp;
//	if ((tmp = strchr(name, INDEXER)) {
//		UNIT x = str2enum(tmp + 1, scfg->unit_levels, U_ILL);
//		if (x == U_ILL) 
//	}
	return set_option(option_struct(name, NULL), value);
}

static inline void set_option_or_die(char *name, const char *value)
{
	epos_option *o = option_struct(name, NULL);
	if (!o) shriek(814, "Unknown option %s", name);
	if (!cfg->langs && (o->structype == OS_LANG || o->structype == OS_VOICE || o->opttype == O_LANG))
		return;

	if (set_option(o, value)) return;
	shriek(814, "Bad value %s for option %s", value, name);
}

/*
 *	For the following one, make sure that base is the correct type
 */

static inline bool set_option(char *name, char *value, void *base, hash_table<char, epos_option> *softopts)
{
	return set_option(option_struct(name, softopts), value, base);
}

bool lang_switch(const char *value)
{
	for (int i=0; i < cfg->n_langs; i++)
		if (!strcmp(cfg->langs[i]->name, value)) {
			if (!cfg->langs[i]->n_voices)
				shriek(462, "Switch to a mute language unimplemented");
			cfg->default_lang = i;
			return true;
		}
	return false;
}

bool voice_switch(const char *value)
{
	cow((cowabilium **)&(this_lang), sizeof(lang), VOICES_OFFSET, VOICES_LENGTH);	//new

	for (int i=0; i < this_lang->n_voices; i++)
		if (!strcmp(this_lang->voices[i]->name, value)) {
			this_lang->default_voice = i;
			return true;
		}
	return false;
}

const char *format_option(epos_option *o, void *base)
{
	char *locus = (char *)base + o->offset;
	switch(o->opttype) {
		case O_BOOL:
			return *(bool *)locus ? "on" : "off";
		case O_MARKUP:
			return enum2str(*(int *)locus, OUT_MLstr);
		case O_SYNTH:
			return enum2str(*(int *)locus, STstr);
		case O_UNIT:
			return enum2str(*(UNIT *)locus, scfg->unit_levels);
		case O_INT:
			sprintf(scratch, "%d", *(int *)locus);
			return scratch;
		case O_STRING: 
			return *(char **)locus;
		case O_CHAR:
			scratch[0] = *(char *)locus;
			scratch[1] = 0;
			return scratch;
		case O_LANG:
			return (char *)this_lang->name;
		case O_VOICE:
			return (char *)this_voice->name;
		case O_CHARSET:
			return enum2str(*(int *)locus, charset_list);
		default: shriek(462, "Bad option type %d", (int)o->opttype);
	}
	return NULL; /* unreachable */
}

const char *format_option(epos_option *o)
{
	switch(o->structype) {
		case OS_STATIC:return format_option(o, scfg);
		case OS_CFG:   return format_option(o, cfg);
		case OS_LANG:  return format_option(o, this_lang);
		case OS_VOICE: return format_option(o, this_voice);
		default: shriek(861, "Bad option class");
	}
	return NULL; /* unreachable */
}

const char *format_option(const char *name)
{
	epos_option *o = option_struct(name, this_lang->soft_options);
	if (!o) {
		shriek(442, "Nonexistent option %s", name);
		return NULL; /* unreachable */
	}
	return format_option(o);
}

int argc_copy = 0;		// these get filled by the args to main()
char **argv_copy = NULL;

void parse_cmd_line()
{
	char *ar;
	char *j;
	register char *par;
	int dees = 0;

	for(int i=1; i<argc_copy; i++) {
		j = ar = argv_copy[i];
		encode_string(ar, cfg->charset, true);
		switch(strspn(ar, CMD_LINE_OPT)) {
		case 3:
			ar+=3;
			if (strchr(ar, CMD_LINE_VAL) && scfg->warnings) 
				shriek(814, "Thrice dashed options have an implicit value");
			set_option_or_die(ar, "0");
			break;
		case 2:
			ar+=2;
			par=strchr(ar, CMD_LINE_VAL);
			if (par) {					//opt=val
				*par=0;
				set_option_or_die(ar, par+1);
				*par=CMD_LINE_VAL;
			} else	if (i+1==argc_copy || strchr(CMD_LINE_OPT, *argv_copy[i+1])) 
					set_option_or_die(ar, "");
				else set_option_or_die(ar, argv_copy[++i]);
			break;
		case 1:
			for (j = ar+1; *j; j++) switch (*j) {
				case 'b': scfg->out_verbose=false; break;
//				case 'd': scfg->show_seg=true; break;
//				case 'e': scfg->show_phones=true; break;
				case 'f': scfg->forking=false; break;
				case 'H': scfg->long_help=true;	/* fall through */
				case 'h': scfg->help=true; break;
				case 'n': cfg->rules_file="nnet.rul";
					  if (cfg->langs && this_lang)
						this_lang->rules_file = cfg->rules_file;
					  scfg->neuronet=true; break;
				case 'p': scfg->pausing=true; break;
				case 's': scfg->play_segs=true; break;
				case 'v': scfg->version=true; break;
				case 'D':
					if (!scfg->use_dbg) scfg->use_dbg=true;
					dees++;
//					else if (scfg->warnings)
//						scfg->always_dbg--;
					break;
				default : shriek(442, "Unknown option -%c, no action taken", *j);
			}
			if (!ar[1]) {
				cfg->input_file = "";   	//dash only
				if (this_lang) this_lang->input_file = "";
			}
			switch (dees) {
				case 0:
					break;
				case 1:
				case 2:
				case 3:
				case 4:
					if (scfg->debug_level > 4 - dees) scfg->debug_level = 4 - dees;
					break;
				default:
					if (scfg->warnings) shriek(814, "Too many dees");
					else break;
			}
			switch (dees) {
				case 0:
					break;
				case 1:
				case 2:
				case 3:
				case 4:
					if (scfg->debug_level > 4 - dees) scfg->debug_level = 4 - dees;
					break;
				default:
					if (scfg->warnings) shriek(814, "Too many dees");
					else break;
			}
			break;
		case 0:
			if (!is_monolith) shriek(814, "Only options allowed at Epos server command line\nUse a client (e.g. \"say\") to specify text");
			if (scfg->input_text && scfg->input_text!=ar) {
				if (!scfg->warnings) break;
				if (cfg->paranoid) shriek(814, "Quotes forgotten on the command line?");
				scratch = (char *) xmalloc(strlen(ar)+strlen(scfg->input_text)+2);
				sprintf(scratch, "%s %s", scfg->input_text, ar);
				ar = FOREVER(strdup(scratch));
				free(scratch);
			}
			scfg->input_text = ar;
			break;
		default:
			if (scfg->warnings) shriek(814, "Too many dashes");
		}
	}
}

void load_config(const char *filename, const char *dirname, const char *what,
		OPT_STRUCT type, void *whither, lang *parent_lang)
{
	int i;

	if (!filename || !*filename) return;
	D_PRINT(3, "Loading %s from %s\n", what, filename);
	char *line = (char *)xmalloc(scfg->max_line + 2) + 2;
	line[-2] = OPT_STRUCT_PREFIX[type];
	line[-1] = ':';
	text *t = new text(filename, dirname, "", what, true);
	while (t->getline(line)) {
		char *value = split_string(line);
		if (value && *value) {
			for (i = strlen(value)-1; strchr(WHITESPACE, value[i]) && i; i--) ;
			if (value[i-1] == ESCAPE && value[i] && i) i++;
			if (value[i]) i++;
			value[i] = 0;		// clumsy: strip off trailing whitespace
		}
		if (!set_option(line - 2, value, whither, parent_lang ? parent_lang->soft_options : (hash_table<char, epos_option> *)NULL)) {
			if (whither == cfg) {		// ...try also static_cfg
				line[-2] = OPT_STRUCT_PREFIX[OS_STATIC];
				if (!set_option(line - 2, value, scfg, (hash_table<char, epos_option> *)NULL))
					shriek(812, "Bad option %s in %s:%d", line, t->current_file, t->current_line);
				line[-2] = OPT_STRUCT_PREFIX[OS_CFG];
			}
		}
	}
	free(line - 2);
	delete t;
}

void load_config(const char *filename)
{
	load_config(filename, scfg->ini_dir, "config", OS_CFG, cfg, NULL);
}

static inline void add_language(const char *lng_name)
{
	char *filename = (char *)xmalloc(strlen(lng_name) + 6);
	char *dirname = (char *)xmalloc(strlen(lng_name) + 6);

	D_PRINT(3, "Adding language %s\n", lng_name);
	sprintf(filename, "%s.ini", lng_name);
	sprintf(dirname, "%s%c%s", scfg->lang_base_dir, SLASH, lng_name);
	if (*lng_name) {
		if (!cfg->langs) cfg->langs = (lang **)xmalloc(8 * sizeof (void *));
		else if (!(cfg->n_langs-1 & cfg->n_langs) && cfg->n_langs > 4)	    // n_langs == 8, 16, 32...
			cfg->langs = (lang **)xrealloc(cfg->langs, (cfg->n_langs << 1) * sizeof(void *));
		cfg->langs[cfg->n_langs++] = new lang(filename, dirname);
	}
	free(filename);
	free(dirname);
}

static inline void load_languages(const char *list)
{
	int i;
	int j=0;
	char *tmp = (char *)xmalloc(strlen(list)+1);


	for (i=0; (tmp[j] = list[i]); i++) {
		if (tmp[j] == ':' ) {
			tmp[j] = 0;
			add_language(tmp);
			j = 0;
		} else j++;
	}
	add_language(tmp);
	free(tmp);
}

void set_base_dir(char *basedir)
{
	scfg->base_dir =  strdup(basedir);
}

static inline void version()
{
	fprintf(cfg->stdshriek, "This is Epos version %s, bug reports to \"%s\" <%s>\n", VERSION, MAINTAINER, MAIL);
}

static inline void dump_help()
{
	int i,j,k;

	printf("usage: %s [options] ['Text to be processed.']\n", argv_copy[0]);
	printf(" -b  bare format (no frills)\n");
//	printf(" -d  show segments\n");
//	printf(" -e  show phones\n");
	printf(" -f  disable forking (detaching) the daemon\n");
	printf(" -n  same as --neuronet --rules_file nnet.rul\n");
	printf(" -p  pausing - show every intermediate state\n");
	printf(" -s  speak it\n");
	printf(" -v  show version\n");
	printf(" -D  debugging info (more D's - more detailed)\n");
	printf(" -   keyboard input\n");
	printf(" --long_options    ...see src/options.lst or 'epos -H' for these\n");
	if (!scfg->long_help) exit(0);

	printf("Long option types: (b) boolean,    (c) character, (e) special enumerated\n");
	printf("                   (n) int number, (s) string,    (u) unit level\n");
	k = 0;
	for (i=0; optlist[i].optname;i++) {
		if (*optlist[i].optname) {
			printf("--%s(%c)", optlist[i].optname, "bueeeencseee"[optlist[i].opttype]);
			for (j = -(signed int)strlen(optlist[i].optname)-5; j <= 0; j += 26) k++;
			for (; j>0; j--) printf(" ");
			if (k >= 3) printf("\n"), k=0;
		}
	}
	printf("\n");
	exit(0);
}

void check_cfg_version(const char *filename)
{
	file *f = claim(filename, "", "", "r", NULL, NULL);
	if (!f) shriek(843, "Configuration files not installed or very old");
	if (strncmp(f->data, VERSION, strlen(VERSION)) && cfg->paranoid) {
		D_PRINT(3, "Expected version %s, found version %s\n",
			VERSION, f->data);
		shriek(843, "Configuration version bad");
	}
	unclaim(f);
}

void config_init()
{
	const char *mlinis[] = {"","ansi.ini","rtf.ini", NULL};

//	if (option_set) shriek(862, "config_init() too late");
	make_option_set();
//	restrict_options();
//	if (privileged_exec())	/* parse restr.ini before command line */
//		restrict_options(); 
	parse_cmd_line();
	restrict_options();
	D_PRINT(3, "Base directory is %s\n", scfg->base_dir);
	D_PRINT(2, "Using configuration file %s\n", scfg->inifile);

	load_config(scfg->fixedfile);
	parse_cmd_line();

	check_cfg_version("version");

	load_config(mlinis[scfg->ml]);
	load_config(scfg->inifile);
	parse_cmd_line();
	load_languages(scfg->languages);

	if (!this_lang->voices || !this_voice) shriek(842, "No voices configured");

	scfg->warnings = true;
	parse_cmd_line();
	scfg->loaded=true;
	
	hash_max_line = scfg->max_line;

	if (scfg->version) version();
	if (scfg->help || scfg->long_help) dump_help();
}

void config_release()
{
	delete option_set;
	option_set = NULL;
}
